#![allow(unused_variables)]

use crate::{
    checked::{Node, NodeKind, Type},
    hash::hash,
    itertewls::Itertewls,
    parsed,
    sym::Syms,
};

pub struct Checker {}

impl Checker {
    pub fn new() -> Self {
        Self {}
    }

    pub fn check(&mut self, ast: &Vec<parsed::Node>) -> Vec<Node> {
        let mut syms = Syms::new();
        ast.into_iter()
            .map(|stmt| self.check_expr(stmt, &mut syms))
            .collect()
    }

    fn fn_scan<'syms>(
        &mut self,
        ast: &Vec<parsed::Node>,
        syms: &mut Syms<'syms>,
    ) -> Result<(), ()> {
        for node in ast {
            match node {
                parsed::Node::Fn {
                    subject,
                    params,
                    return_typ,
                    body,
                } => {
                    let params = params.into_iter().map(|param| {
                        let parsed::Node::Param { subject, typ } = param else { unreachable!() };
                        let parsed::Node::Id(id) = subject.as_ref() else { unreachable!() };
                        let typ = self.check_type(typ.as_ref().ok_or(())?);
                        Ok((*id, self.node(NodeKind::Param { subject: Box::new(self.node(NodeKind::Id(*id), Type::Unit)), typ: Some(typ) }, Type::Unit)))
                    }).collect::<Result<Vec<_>, _>>()?;

                    if let Some(id) = params.iter().map(|(id, _)| *id).find_first_duplicate() {
                        self.error("redefinition param");
                        return Err(());
                    }

                    let parsed::Node::Id(id) = subject.as_ref() else { unreachable!() };

                    if syms.defined_locally(*id) {
                        self.error("redefinition fn");
                        return Err(());
                    }
                    //syms.define(*id, typ.clone());
                }
                _ => {}
            }
        }
        Ok(())
    }

    fn check_expr<'syms>(&mut self, node: &parsed::Node, syms: &mut Syms<'syms>) -> Node {
        match node {
            parsed::Node::Error => self.node(NodeKind::Error, Type::Unit),
            parsed::Node::Id(id) => {
                let Some(sym) = syms.get(*id) else {
                    self.error("undefined >~<");
                    return self.node(NodeKind::Error, Type::Error);
                };
                self.node(NodeKind::Id(*id), sym.typ)
            }
            parsed::Node::Int(value) => self.node(
                NodeKind::Int(*value),
                if *value as i32 > i32::MAX {
                    Type::U32
                } else {
                    Type::I32
                },
            ),
            parsed::Node::String(value) => self.node(NodeKind::String(value.clone()), Type::String),
            parsed::Node::Group(expr) => {
                let expr = self.check_expr(expr, syms);
                let typ = expr.typ.clone();
                self.node(NodeKind::Group(Box::new(expr)), typ)
            }
            parsed::Node::Block(stmts) => {
                let mut child_syms = syms.child();
                let stmts = stmts
                    .into_iter()
                    .map(|stmt| self.check_expr(stmt, &mut child_syms))
                    .collect::<Vec<_>>();
                let typ = stmts
                    .last()
                    .map(|stmt| stmt.typ.clone())
                    .unwrap_or(Type::Unit);
                self.node(NodeKind::Block(stmts), typ)
            }
            parsed::Node::Call { subject, args } => {
                let subject = Box::new(self.check_expr(subject, syms));

                let args = args
                    .into_iter()
                    .map(|arg| self.check_expr(arg, syms))
                    .collect::<Vec<_>>();
                let typ = 'br: {
                    match subject.typ.clone() {
                        Type::Fn {
                            id: _,
                            params,
                            return_typ,
                        } => {
                            if args.len() != params.len() {
                                self.error("too few/many args");
                                break 'br Type::Error;
                            }
                            if args
                                .iter()
                                .zip(params)
                                .map(|(arg, param)| self.compatible(&arg.typ, &param.typ))
                                .any(|is_compatible| !is_compatible)
                            {
                                self.error("incorrect args");
                                break 'br Type::Error;
                            }
                            *return_typ
                        }
                        _ => {
                            self.error("not a function");
                            Type::Error
                        }
                    }
                };
                self.node(NodeKind::Call { subject, args }, typ)
            }
            parsed::Node::If {
                cond,
                truthy,
                falsy,
            } => {
                let cond = Box::new(self.check_expr(cond, syms));
                let truthy = Box::new(self.check_expr(truthy, syms));
                let falsy = falsy
                    .as_ref()
                    .map(|block| Box::new(self.check_expr(block, syms)));
                let typ = 'br: {
                    match falsy.as_ref().map(|block| block.typ.clone()) {
                        Some(falsy_typ) => {
                            if !self.compatible(&truthy.typ, &falsy_typ) {
                                self.error("incompatible types #2");
                                break 'br Type::Error;
                            }
                            falsy_typ
                        }
                        None => Type::Unit,
                    }
                };
                self.node(
                    NodeKind::If {
                        cond,
                        truthy,
                        falsy,
                    },
                    typ,
                )
            }
            parsed::Node::Loop { body } => {
                let body = Box::new(self.check_expr(body, &mut syms.child()));
                let typ = body.typ.clone();
                self.node(NodeKind::Loop { body }, typ)
            }
            parsed::Node::Break => self.node(NodeKind::Break, Type::Unit),
            parsed::Node::Assign { subject, value } => {
                let subject = Box::new(self.check_expr(subject, syms));
                let value = Box::new(self.check_expr(value, syms));

                let typ = if !self.compatible(&subject.typ, &value.typ) {
                    self.error("incompatible types #3");
                    Type::Error
                } else {
                    subject.typ.clone()
                };
                self.node(NodeKind::Assign { subject, value }, typ)
            }
            parsed::Node::Let { subject, value } => {
                let (subject, subject_typ) = match subject.as_ref() {
                    parsed::Node::Param { subject, typ } => {
                        (subject, typ.as_ref().map(|typ| self.check_type(typ)))
                    }
                    _ => unreachable!(),
                };

                let value = Box::new(self.check_expr(value, syms));
                let typ = value.typ.clone();

                if subject_typ
                    .as_ref()
                    .map(|subject_typ| !self.compatible(subject_typ, &typ))
                    .unwrap_or(false)
                {
                    self.error("incompatible types #1");
                    return self.node(NodeKind::Error, Type::Error);
                }

                let subject = match subject.as_ref() {
                    parsed::Node::Id(id) => {
                        if syms.defined_locally(*id) {
                            self.error("redefinition");
                            return self.node(NodeKind::Error, Type::Error);
                        }
                        syms.define(*id, typ.clone());
                        Box::new(self.node(
                            NodeKind::Param {
                                subject: Box::new(self.node(NodeKind::Id(*id), Type::Unit)),
                                typ: Some(Type::Unit),
                            },
                            Type::Unit,
                        ))
                    }
                    _ => unreachable!(),
                };

                self.node(NodeKind::Let { subject, value }, typ)
            }
            parsed::Node::Fn {
                subject,
                params,
                return_typ,
                body,
            } => {
                todo!("symbol lookup");
            }
            parsed::Node::Return { value } => {
                let value = value
                    .as_ref()
                    .map(|value| Box::new(self.check_expr(value, syms)));
                let typ = value
                    .as_ref()
                    .map(|value| value.typ.clone())
                    .unwrap_or(Type::Unit);
                self.node(NodeKind::Return { value }, typ)
            }
            parsed::Node::Param { .. } => unreachable!("handle elsewhere"),
        }
    }

    fn check_type(&self, node: &parsed::Node) -> Type {
        match node {
            parsed::Node::Error => Type::Error,
            parsed::Node::Id(value) => {
                if *value == hash("i32") {
                    Type::I32
                } else if *value == hash("u32") {
                    Type::U32
                } else {
                    todo!("symbol lookup idk")
                }
            }
            _ => unreachable!(),
        }
    }

    fn compatible(&self, typ_a: &Type, typ_b: &Type) -> bool {
        typ_a == typ_b
    }

    fn node(&self, kind: NodeKind, typ: Type) -> Node {
        Node { kind, typ }
    }

    fn error<S: Into<String>>(&mut self, msg: S) {
        let msg = msg.into();
        println!("checker error: {msg}");
    }
}

#[test]
fn test_checker() {
    use crate::parser::Parser;
    use NodeKind::{Block, Id, Int, Let, Param};
    use Type::{Unit, I32};

    let check = |text| Checker::new().check(&Parser::new(text).parse());

    assert_eq!(
        check("let a = 5; a;"),
        vec![
            Node {
                kind: Let {
                    subject: Box::new(Node {
                        kind: Param {
                            subject: Box::new(Node {
                                kind: Id(hash("a")),
                                typ: Unit
                            }),
                            typ: Some(Unit)
                        },
                        typ: Unit
                    }),
                    value: Box::new(Node {
                        kind: Int(5),
                        typ: I32
                    })
                },
                typ: I32
            },
            Node {
                kind: Id(hash("a")),
                typ: I32
            }
        ]
    );

    assert_eq!(
        check("let a = 5; a; { a; let b = 5; b; } a; b;"),
        vec![
            Node {
                kind: Let {
                    subject: Box::new(Node {
                        kind: Param {
                            subject: Box::new(Node {
                                kind: Id(hash("a")),
                                typ: Unit
                            }),
                            typ: Some(Unit)
                        },
                        typ: Unit
                    }),
                    value: Box::new(Node {
                        kind: Int(5),
                        typ: I32
                    })
                },
                typ: I32
            },
            Node {
                kind: Id(hash("a")),
                typ: I32
            },
            Node {
                kind: Block(vec![
                    Node {
                        kind: Id(hash("a")),
                        typ: I32
                    },
                    Node {
                        kind: Let {
                            subject: Box::new(Node {
                                kind: Param {
                                    subject: Box::new(Node {
                                        kind: Id(hash("b")),
                                        typ: Unit
                                    }),
                                    typ: Some(Unit)
                                },
                                typ: Unit
                            }),
                            value: Box::new(Node {
                                kind: Int(5),
                                typ: I32
                            })
                        },
                        typ: I32
                    },
                    Node {
                        kind: Id(hash("b")),
                        typ: I32
                    }
                ]),
                typ: I32
            },
            Node {
                kind: Id(hash("a")),
                typ: I32
            },
            Node {
                kind: NodeKind::Error,
                typ: Type::Error
            }
        ]
    );
}
